//! this example demonstrates the type of four-core clocked inverter conceived
//! of in the https://uninsane.org/ blogpost.
//!
//! we're going for something like this:
//!
//! ```
//!          ____         ____         ____         ____
//!  + ___  /    \ +___  /    \ +___  /    \ +___  /    \  ___ +
//! IN ___ |o    o| ___ |x    o| ___ |x    o| ___ |x    o| ___ OUT
//!  -      \_o__/ -     \_o__/ -     \_o__/ -     \_o__/      -
//!           ||           ||           ||           ||
//!          +  -         +  -         +  -         +  -
//!          CTL0         CTL1         CTL2         CTL3
//! ```
//!
//! conventions:
//! - CW core polarization is logic '1'.
//! - differential signals are assumed to flow clockwise. i.e. left-to-right, with + on top and -
//!   on bottom.
//! - 'x' denotes a signal going "into" the page. 'o' for a signal going "out of" the page.
//! - the core driven by CTLn is named Sn, interchangeably used to represent the state (0 or 1) of
//!   that core
//!
//! hence:
//! - the input into core S0 has current flowing out of the page and tends to drive S0 to logic '0'.
//! - the input into S1, S2, S3 tends these cores to logic '1'.
//! - a positive current to CTLn will "clear" Sn to the logic '0' state.
//!
//! in this device, S0 is effectively an inverter, with S1, S2, S3 acting as buffers.
//! for the purpose of this simulation, all terminal wires are closed loops either explicitly
//! driven or measured.
//!
//! to run this, from toplevel directory:
//! ```
//! $ cargo run --release --bin multi_core_inverter
//! $ pushd crates/coremem; cargo run --release --bin viewer ../../out/applications/multi_core_inverter/0/ ; popd
//! ```

use coremem::geom::{Coord as _, Meters, Torus};
use coremem::mat::{Ferroxcube3R1MH, IsoConductorOr, IsomorphicConductor};
use coremem::meas;
use coremem::real::{self, Real as _};
use coremem::sim::spirv::{self, SpirvSim};
use coremem::sim::units::Seconds;
use coremem::stim::{CurlStimulus, Exp, Gated, Sinusoid1, TimeVarying as _};
use coremem::Driver;

type R = real::R32;
type Mat = IsoConductorOr<R, Ferroxcube3R1MH>;
// type Backend = spirv::CpuBackend;
type Backend = spirv::WgpuBackend;
type Sim = SpirvSim::<R, Mat, Backend>;

#[derive(Clone, Copy)]
enum DriveType {
    HoldHigh,
    ReleaseHigh,
    HoldLow,
    ReleaseLow,
    Float,
}

impl DriveType {
    fn direction(&self) -> i32 {
        use DriveType::*;
        match self {
            HoldHigh | ReleaseHigh => 1,
            HoldLow | ReleaseLow => -1,
            Float => 0,
        }
    }
    fn is_hold(&self) -> bool {
        match self {
            Self::HoldHigh | Self::HoldLow => true,
            _ => false,
        }
    }
    fn is_release(&self) -> bool {
        match self {
            Self::ReleaseHigh | Self::ReleaseLow => true,
            _ => false,
        }
    }
}

fn main() {
    coremem::init_logging();
    // coremem::init_debug();
    let um = |n| n as f32 * 1e-6;
    // let ns = |n| n as f32 * 1e-9;
    let ps = |n| n as f32 * 1e-12;
    let feat_size = um(10);

    let input_magnitude = 1.0e9;
    let clock_phase_duration = ps(2000);

    let s_major = um(160);
    let s_minor = um(30);
    let io_major = um(80);
    let io_minor = um(30);
    let coupling_major = um(130);
    let coupling_minor = um(30);

    let sx = |n| um((n+1) * 400);
    let sy = um(400);
    let sz = um(280);
    let couplingx = |n| sx(n) + um(200);

    let sim_bounds = Meters::new(sx(4), sy * 2.0, sz * 2.0);
    let sim_padding = Meters::new(um(80), um(80), um(80));

    let drive0 = Torus::new_xz(Meters::new(sx(0) - s_major, sy, sz), io_major, io_minor);
    let sense3 = Torus::new_xz(Meters::new(sx(3) + s_major, sy, sz), io_major, io_minor);
    let ctl = |n| Torus::new_yz(Meters::new(sx(n), sy + s_major, sz), io_major, io_minor);
    let s = |n| Torus::new_xy(Meters::new(sx(n), sy, sz), s_major, s_minor);
    // coupling(n) is the wire which couples core n into core n+1
    let coupling = |n| Torus::new_xz(Meters::new(couplingx(n), sy, sz), coupling_major, coupling_minor);

    let control_signal = |driver: &mut Driver<Sim>, region: &Torus, cycle: u32, ty: DriveType| {
        let area = region.cross_section();
        // negation because blog confuses current directions.
        let amp = -ty.direction() as f32 * input_magnitude / area;
        let start = clock_phase_duration * cycle as f32;
        if ty.is_hold() {
            // simple square wave:
            let wave = Gated::new(amp, start, start + clock_phase_duration);
            let stim = CurlStimulus::new(
                region.clone(),
                wave.clone(),
            );
            driver.add_stimulus(stim);
        } else if ty.is_release() {
            // decaying exponential wave:
            let wave = Exp::new_at(amp, start, 0.125 * clock_phase_duration /* half-life */);
            let stim = CurlStimulus::new(
                region.clone(),
                wave.clone(),
            );
            driver.add_stimulus(stim);
        } // else: Float
        // sine wave (untested):
        // let wave = Sinusoid1::from_wavelength(direction as f32 * input_magnitude / area, clock_phase_duration * 2.0)
        //     .half_cycle()
        //     .shifted(clock_phase_duration * cycle as f32);
    };

    let wire_mat = IsomorphicConductor::new(1e6f32.cast::<R>());
    // let ferro_mat = wire_mat;
    let ferro_mat = Ferroxcube3R1MH::new();

    let mut driver = Driver::new(Sim::new(
        sim_bounds.to_index(feat_size), feat_size,
    ));
    driver.add_classical_boundary_explicit::<R, _>(sim_padding);

    //////// create the wires and toroids
    driver.fill_region(&drive0, wire_mat);
    driver.fill_region(&sense3, wire_mat);
    for core in 0..4 {
        driver.fill_region(&s(core), ferro_mat);
        driver.fill_region(&ctl(core), wire_mat);
        if core != 3 {
            driver.fill_region(&coupling(core), wire_mat);
        }
    }

    //////// monitor some measurements
    // driver.add_measurement(meas::CurrentLoop::new("drv0", drive0.clone()));
    for core in 0..4 {
        driver.add_measurement(meas::CurrentLoop::new(
            &format!("drive{}", core),
            ctl(core),
        ));
    }
    for core in 0..3 {
        driver.add_measurement(meas::CurrentLoop::new(
            &format!("sense{}", core),
            coupling(core),
        ));
    }
    for core in 0..4 {
        driver.add_measurement(meas::MagneticLoop::new(
            &format!("state{}", core),
            s(core),
        ));
    }
    driver.add_measurement(meas::CurrentLoop::new("sense3", sense3.clone()));

    //////// create the stimuli

    // each row N denotes the drive currents at clock cycle N.
    // each col M denotes the drive current at core M.
    // this drive map is taken from the blog article as of 2022-08-09; it has noted errors
    // let drive_map = [
    //     [1, 0, 1, 1],
    //     [1, 1, 0, 1],
    //     [1, 1, 1, 0],
    //     [0, 1, 1, 1i32],
    // ];
    // proposed drive map: the location where a core is charged to 1 requires a negative current.
    // let drive_map = [
    //     [ 1,  0,  1,  1],    // clear S0 -> S1; hold S2, S3
    //     [-1,  1,  0,  1],    // clear S1 -> S2; hold S3; reset S0
    //     [-1,  1,  1,  0],    // clear S2 -> S3; hold S0 (high), S1
    //     [ 0,  1, -1,  1i32], // clear S3 -> Sn; hold S1, S2 (S0 gets new input
    // ];
    // the above drive map has some implicit requirement of clock lag. it won't work with perfectly
    // square clocks. here's a realizable drive map:
    // #[allow(non_snake_case)]
    // let F = 0; // indicates a gate which has been driven to a known value and is now floating
    // #[allow(non_snake_case)]
    // let D = 0; // indicates a gate which carries data.
    // let drive_map = [
    //     [-1,  1,  1,  1],    // initial state: S0=1, S1=S2=S3=0
    //     [-1,  F,  1,  1],    // open S1 for write

    //     [ 1,  D,  1,  1],    // clear S0 -> S1; hold S2, S3
    //     [ 1,  D,  F,  1],    // open S2 for write

    //     [-1,  1,  D,  1],    // clear S1 -> S2; hold S3; reset S0
    //     [-1,  1,  D,  F],    // open S3 for write

    //     [-1,  1,  1,  D],    // clear S2 -> S3; hold S0 (high), S1
    //     [ F,  1,  1,  D],    // open S0 for write (not used in this demo)

    //     [ D,  1,  1,  1i32], // clear S3 -> Sn; hold S1, S2 (S0 gets new input)
    //     [ D,  F,  1,  1i32], // open S1 for write (not used; for completeness)
    // ];

    use DriveType::*;
    let drive_map = [
        [HoldLow,     HoldHigh,     HoldHigh,     HoldHigh],    // initial state: S0=1, S1=S2=S3=0
        [HoldLow,     ReleaseHigh,  HoldHigh,     HoldHigh],    // open S1 for write

        [HoldHigh,    Float,        HoldHigh,     HoldHigh],    // clear S0 -> S1; hold S2, S3
        [HoldHigh,    Float,        ReleaseHigh,  HoldHigh],    // open S2 for write

        [HoldHigh,    HoldHigh,     Float,        HoldHigh],    // clear S1 -> S2; hold S0, S3
        // NB: bonus clock! we have to reset S0 before opening S3, else writeback from next stage
        [HoldLow,     HoldHigh,     Float,        HoldHigh],    // reset S0; hold S1, S3
        [HoldLow,     HoldHigh,     Float,        ReleaseHigh], // open S3 for write

        [HoldLow,     HoldHigh,     HoldHigh,     Float],       // clear S2 -> S3; hold S0 (high), S1
        [ReleaseLow,  HoldHigh,     HoldHigh,     Float],       // open S0 for write (not used in this demo)

        [Float,       HoldHigh,     HoldHigh,     HoldHigh],    // clear S3 -> Sn; hold S1, S2 (S0 gets new input)
        [Float,       ReleaseHigh,  HoldHigh,     HoldHigh],    // open S1 for write (not used; for completeness)
    ];

    let cycles = drive_map.len() as u32;
    let duration = Seconds(clock_phase_duration * (cycles + 3) as f32);

    for cycle in 0..cycles {
        for core in 0..4 {
            // current polarity of the drive wires is inverted in this model v.s. the blog article,
            // so invert our currents in order to achieve the magnetic states of the article.
            let sig = drive_map[cycle as usize][core as usize];
            control_signal(&mut driver, &ctl(core), cycle, sig);
        }
    }

    let prefix = "out/applications/multi_core_inverter/8/";
    let _ = std::fs::create_dir_all(&prefix);
    // driver.add_state_file(&*format!("{}state.bc", prefix), 9600);
    driver.add_serializer_renderer(&*format!("{}frame-", prefix), 400, None);
    driver.add_csv_renderer(&*format!("{}meas-detailed.csv", prefix), 100, None);
    driver.add_csv_renderer(&*format!("{}meas.csv", prefix), 800, None);
    driver.set_steps_per_stim(100);

    driver.step_until(duration);
}
